import json
import logging
import os
from collections import UserDict
from pathlib import Path
from urllib.parse import urlencode

from puddl.apps import App

log = logging.getLogger(__name__)


class ConfigError(Exception):
    pass


class Config(UserDict):
    PUDDLRC = Path.home() / '.puddlrc'

    DB_DEFAULTS = {
        'PGUSER': os.environ.get('PGUSER'),
        'PGPASSWORD': os.environ.get('PGPASSWORD'),
        'PGHOST': os.environ.get('PGHOST'),
        'PGPORT': os.environ.get('PGPORT'),
        'PGDATABASE': os.environ.get('PGDATABASE'),
        'PGAPPNAME': os.environ.get('PGAPPNAME', 'puddl'),
    }

    @classmethod
    def get_defaults(cls):
        defaults = cls.DB_DEFAULTS.copy()
        defaults['apps'] = []
        return defaults

    def write(self):
        # serialize first, write next
        x = json.dumps(self.data, indent=2, sort_keys=True)
        with self.PUDDLRC.open('w') as f:
            f.write(x)

    @property
    def apps(self):
        return self['apps']

    @property
    def db_url(self) -> str:
        db_params = {
            'application_name': self['PGAPPNAME']
        }
        return 'postgresql://{PGUSER}:{PGPASSWORD}@{PGHOST}:{PGPORT}' \
               '/{PGDATABASE}?{params}'.format(**self, params=urlencode(db_params))

    @property
    def db_env(self):
        return {k: v for k, v in self.items() if k in self.DB_DEFAULTS.keys()}

    def add_app(self, name):
        p = App(name)
        p.validate()
        self['apps'].append(p.pkg_name)

    def remove_app(self, name):
        ps = self['apps']
        try:
            del ps[ps.index(name)]
        except ValueError:
            log.warning(f'app {name} was not installed')


def get_config(app=None):
    if app is None:
        app = 'puddl'

    try:
        with Config.PUDDLRC.open() as f:
            data = json.load(f)
    except FileNotFoundError:
        data = Config.get_defaults()
    except json.decoder.JSONDecodeError as e:
        raise ConfigError(
            f'{Config.PUDDLRC}:{e.lineno} column {e.colno} char {e.pos}: {e.msg}'
        )
    data['PGAPPNAME'] = app
    return Config(**data)



from itertools import chain

from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import CreateColumn

from puddl import get_config


def get_alchemy_session(conf=None, **kwargs):
    """
    Convenience. Don't know if this is wise. :>
    """
    if conf is None:
        conf = get_config()
    import sqlalchemy
    from sqlalchemy.orm import sessionmaker
    engine = sqlalchemy.create_engine(conf.db_url, **kwargs)

    Session = sessionmaker(bind=engine)
    return Session()


# https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#postgresql-10-identity-columns
@compiles(CreateColumn, 'postgresql')
def use_identity(element, compiler, **kw):
    text = compiler.visit_create_column(element, **kw)
    text = text.replace("SERIAL", "INT GENERATED BY DEFAULT AS IDENTITY")
    return text


def make_base(schema):
    class Base:
        __table_args__ = {"schema": schema}

    return declarative_base(cls=Base)


class Upserter:
    def __init__(self, session, model):
        self.session = session
        self.model = model
        inspector = Inspector.from_engine(session.bind)
        _xs = inspector.get_unique_constraints(self.model.__tablename__,
                                               schema=self.model.__table__.schema)
        list_of_lists = [x['column_names'] for x in _xs]
        self.unique_columns = set(chain.from_iterable(list_of_lists))

    def upsert(self, **data):
        # read this carefully:
        # https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#sqlalchemy.dialects.postgresql.dml.Insert.on_conflict_do_update
        columns = set(data.keys())
        update_columns = columns - self.unique_columns
        insert_stmt = insert(self.model).values(**data)
        excluded_mapping = {  # LHS: column name, RHS: Column instance
            col: getattr(insert_stmt.excluded, col)
            for col in update_columns
        }
        do_update_stmt = insert_stmt.on_conflict_do_update(
            index_elements=self.unique_columns,
            set_=excluded_mapping,
        )
        self.session.execute(do_update_stmt)



import logging
import subprocess

import click
from pathlib import Path

from puddl import get_config
from puddl.apps import App
from puddl.conf import Config

log = logging.getLogger(__name__)
LOG_LEVELS = ['CRITICAL', 'FATAL', 'ERROR', 'WARN', 'WARNING', 'INFO', 'DEBUG']


class AppsGroup(click.Group):

    def __init__(self, name=None, commands=None, **attrs):
        self._apps = [App(x) for x in get_config().apps]
        self._name2app = {
            x.cmd_name: x for x in self._apps
        }
        super().__init__(name, commands, **attrs)

    def list_commands(self, ctx):
        core_commands = set(super().list_commands(ctx))
        app_names = set(x for x in self._name2app)
        return sorted(core_commands | app_names)

    def get_command(self, ctx, cmd_name):
        if cmd_name in self._name2app:
            app = self._name2app[cmd_name]
            return app.get_command()
        return super().get_command(ctx, cmd_name)


@click.group(cls=AppsGroup)
@click.option('-l', '--log-level',
              type=click.Choice(LOG_LEVELS, case_sensitive=False),
              default='WARNING')
@click.option('--loggers', default='', help='comma-separated logger names')
@click.version_option()
def root(log_level, loggers):
    logging.basicConfig(level=log_level)
    if loggers:
        for name in loggers.split(','):
            logging.getLogger(name).setLevel(log_level)
    pass


@root.group()
def config():
    pass


@config.command()
def init():
    data = {}
    defaults = Config.get_defaults()
    data.update(defaults)

    dot_env = Path('.env')
    try:
        with dot_env.open() as f:
            contents = f.read()
        for line in contents.strip().split('\n'):
            k, v = line.strip().split('=')
            if k in defaults:
                log.info(f'using {k} from .env')
                data[k] = v
    except FileNotFoundError:
        pass

    conf = Config(data)
    conf.write()


@config.command()
def show():
    conf = get_config()
    print(conf.data)


@root.group()
def app():
    pass


@app.command('add')
@click.argument('package_path')
def app_add(package_path):
    conf = get_config()
    conf.add_app(package_path)
    conf.write()


@app.command('ls')
@click.option('-v', '--verbose', is_flag=True)
def app_ls(verbose):
    conf = get_config()
    apps = sorted(conf['apps'])
    for app_name in apps:
        msg = app_name
        if verbose:
            msg = str(App(app_name))
        click.echo(msg)


@app.command('rm')
@click.argument('package_path')
def app_remove(package_path):
    conf = get_config()
    conf.remove_app(package_path)
    conf.write()


@root.group()
def db():
    pass


@db.command()
def health():
    conf = get_config()

    import psycopg2
    # consumes PG* env vars
    connection = psycopg2.connect(conf.db_url)
    with connection.cursor() as cursor:
        cursor.execute('SELECT 1')
    log.info('database available')


@db.command()
def shell():
    conf = get_config(app='db shell')
    # want to see good error handling for this kind of thing?
    # https://github.com/pallets/click/blob/19fdc8509a1b946c088512e0e5e388a8d012f0ce/src/click/_termui_impl.py#L472-L487
    subprocess.Popen('psql', env=conf.db_env).wait()


@db.command()
@click.option('--app')
def url(app):
    """
    print DB URL

    Useful for things like this:

        from sqlalchemy import create_engine
        database_url = 'postgresql://puddl:aey1Ki4oaZohseWod2ri@localhost:13370/puddl'
        engine = create_engine(database_url, echo=False)
        df.to_sql('bp', engine)
    """
    conf = get_config(app=app)
    print(conf.db_url)


@db.command()
def env():
    """
    Prints the database's environment.
    WARNING: This dumps your password. Use it like this:

        source <(puddl db env)
    """
    conf = get_config()
    for k, v in conf.db_env.items():
        print(f'export {k}={v}')


@db.command()
def sessions():
    """
    Lists active sessions.
    """
    conf = get_config(app='db sessions')
    query = """SELECT pid AS process_id,
           usename AS username,
           datname AS database_name,
           client_addr AS client_address,
           application_name,
           backend_start,
           state,
           state_change
    FROM pg_stat_activity;"""
    click.echo(
        subprocess.check_output(['psql', '-c', query], encoding='utf-8', env=conf.db_env))


@db.command()
def queries():
    print('xxx')


def main():
    # this function is referenced in `setup.py`
    root(auto_envvar_prefix='PUDDL')


if __name__ == '__main__':
    # https://click.palletsprojects.com/en/7.x/options/#values-from-environment-variables
    main()



# Generated by Django 3.0.4 on 2020-03-15 11:05

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import puddl.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('puddl', '0002_device'),
        ('file', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('path', puddl.fields.PathField()),
                ('filetype', django.contrib.postgres.fields.jsonb.JSONField(null=True)),
                ('stat', django.contrib.postgres.fields.jsonb.JSONField()),
                ('device', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='puddl.Device')),
            ],
        ),
        migrations.AddConstraint(
            model_name='file',
            constraint=models.UniqueConstraint(
                fields=('device', 'path'), name='unique_device_path'),
        ),
    ]
